Set Implicit Arguments.
Require Import LibTactics.
Require Import CFLib.

(* The stuff in this file should possibly move to CFML. *)

(* -------------------------------------------------------------------------- *)

(* This type class allows rewriting (using [setoid_rewrite]) under an [Hexists]
   quantifier. *)

Require Import Coq.Classes.Morphisms.
Require Import LibAxioms.

Program Instance :
  forall A, Proper (pointwise_relation A eq ==> eq) (@heap_is_pack A).
Next Obligation.
  intros Q1 Q2 h. f_equal. apply func_ext_dep. intros. rewrite h. reflexivity.
Qed.

Goal
  forall A (P Q : A -> hprop),
  (forall a, P a = Q a) ->
  (Hexists a, P a) ==> (Hexists a, Q a).
Proof.
  introv h. setoid_rewrite h. hsimpl.
Qed.

(* -------------------------------------------------------------------------- *)

(* Tactics. *)

Lemma heap_weaken_pack_post : forall A B (x:A) Q (J : B -> A -> hprop),
  Q ===> (fun y => J y x) -> Q ===> (fun y => heap_is_pack (J y)).
Proof using. introv W h. exists x. apply~ W. Qed.

Ltac hhsimpl := (* TEMPORARY *)
  repeat instantiate (1 := \[]); hsimpl.

Tactic Notation "hexists" :=
  eapply heap_weaken_pack.
(* TEMPORARY use [xsimpl] instead *)
(* TODO: clarify if there is redundancy with [xgen]
   by the way, [xgen] is undocumented *)

Tactic Notation "xexists" :=
  xpost; [ | eapply heap_weaken_pack_post; eauto ].

Tactic Notation "hexists" constr(witness) :=
  eapply heap_weaken_pack with (x := witness).

Tactic Notation "hexists" constr(witness1) constr(witness2) :=
  hexists witness1; hexists witness2.

Tactic Notation "xexists" constr(witness) :=
  xpost; [ | eapply heap_weaken_pack_post with (x := witness); eauto  ].

Tactic Notation "xexists" constr(witness1) constr(witness2) :=
  xexists witness1; xexists witness2.

(* Paraphrase lemma. Unused at the moment; could be useful? *)

Lemma unfold_hdata:
  forall (A : Type) (P : A -> hprop),
  hdata P = P.
Proof.
  reflexivity.
Qed.

Lemma unfold_tilde:
  forall {A : Type} (Q : A -> hprop),
  (fun x : A => x ~> Q) = Q.
Proof.
  reflexivity.
Qed.

(* -------------------------------------------------------------------------- *)

(* Lemmas. *)

Lemma prove_pure_consequence: (* TEMPORARY move to CFHeaps, if not already there *)
  forall (P : hprop) (Q : Prop),
  (forall h, P h -> Q) ->
  P ==> P \* \[ Q ].
Proof.
  introv H. intros h HPh.
  exists h. exists heap_empty. splits.
  { eauto. }
  { eapply heap_is_empty_st_prove. eauto. }
  { eapply heap_disjoint_empty_r. }
  { rewrite heap_union_neutral_r. reflexivity. }
Qed.

(* -------------------------------------------------------------------------- *)

(* A notion of view shift. *)

(* TEMPORARY maybe the definition of view shift should be:
   Ret tt. *)

Definition view_shift (P1 P2 : hprop) : Prop :=
  forall (B : Type) (F : ~~B) (Q : B -> hprop),
  is_local F ->
  F P2 Q ->
  F P1 Q.

Lemma implication_as_view_shift:
  forall P1 P2,
  P1 ==> P2 ->
  view_shift P1 P2.
Proof.
  introv H Hlocal Hspec.
  xchange H.
  rewrite star_neutral_r.
  assumption.
Qed.

Lemma implication_with_slack_as_view_shift:
  forall P1 P2 slack,
  P1 ==> P2 \* slack ->
  view_shift P1 P2.
Proof.
  introv H Hlocal Hspec.
  xchange H. xgc slack.
  rewrite star_neutral_r.
  assumption.
Qed.

Lemma view_shift_weaken_pre:
  forall P1 P2 Q,
  view_shift P2 Q ->
  P1 ==> P2 ->
  view_shift P1 Q.
Proof.
  unfold view_shift. introv Hview Hsub. intros B F Q'. intros.
  eapply local_weaken_pre; [ eauto | | eexact Hsub ].
  eapply Hview; eauto.
Qed.

Lemma view_shift_weaken_post:
  forall P Q1 Q2,
  view_shift P Q1 ->
  Q1 ==> Q2 ->
  view_shift P Q2.
Proof.
  unfold view_shift. introv Hview Hsub. intros B F Q'. intros.
  eapply Hview. eauto.
  eapply local_weaken_pre; eauto.
Qed.

Lemma view_shift_gc_pre:
  forall P1 P2 Q,
  view_shift P1 Q ->
  view_shift (P1  \* P2) Q.
Proof.
  unfold view_shift. introv Hview. intros.
  eapply local_gc_pre. eauto. instantiate (1 := P1). hsimpl.
  eapply local_weaken. eauto. eapply Hview; eauto. hsimpl. hsimpl.
Qed.

Lemma is_local_frame:
  forall B (F : ~~B) R,
  is_local F ->
  is_local (fun P Q => F (P \* R) Q).
Proof.
  introv Hlocal.
  unfold is_local, local. extens; intros P Q. split.
Abort. (* TEMPORARY big problem! *)

Lemma view_shift_frame:
  forall P Q R,
  view_shift P Q ->
  view_shift (P \* R) (Q \* R).
Proof.
  unfold view_shift. introv Hview. intros B F Q'. intros.
  eapply Hview with (F := fun P Q => F (P \* R) Q).
Admitted. (* TEMPORARY big problem!
    { eauto using is_local_frame. }
    { eauto. }
Qed. *)

Lemma apply_view_shift:
  forall (H H2 P1 P2 : hprop) (B : Type) (F : ~~B) (Q : B -> hprop),
  is_local F ->
  view_shift P1 P2 ->
  H ==> P1 \* H2 ->
  F (P2 \* H2) Q ->
  F H Q.
Proof.
  introv Hlocal Hview ? ?.
  eapply view_shift_weaken_pre; [ | eauto | eauto | eauto ].
  eapply view_shift_frame.
  eauto.
Qed.

Tactic Notation "view_shift" constr(H) :=
  eapply apply_view_shift; [ xlocal | eapply H | hsimpl | ].

Lemma apply_view_shift_afterwards:
  forall (H : hprop) (B : Type) (F : ~~B) (Q1 Q2 : B -> hprop),
  is_local F ->
  F H Q1 ->
  (forall y, view_shift (Q1 y) (Q2 y)) ->
  F H Q2.
Proof.
  unfold view_shift.
  intros.
  (* TEMPORARY Intuitively, F is equivalent to "Let x := F in Ret x"
     so applying the view shift *after* F is the same as applying it *before*
     returning the result. *)
Admitted.

Tactic Notation "view_shift_afterwards" constr(H) :=
  eapply apply_view_shift_afterwards; [ xlocal | | eapply H ].

Lemma view_shift_transitive:
  forall P Q R,
  view_shift P Q ->
  view_shift Q R ->
  view_shift P R.
Proof.
  introv H1 H2 Hlocal Hview.
  view_shift H1.
  view_shift H2.
  rewrite star_neutral_r.
  assumption.
Qed.

Lemma view_shift_exists:
  forall X (P Q : X -> hprop),
  (forall x, view_shift (P x) (Q x)) ->
  view_shift (heap_is_pack P) (heap_is_pack Q).
Proof.
  introv Hview. introv Hlocal Hspec.
  xpull. intros x.
  eapply Hview; [ eauto | ].
  eapply local_weaken_pre; [ eauto | eapply Hspec | hsimpl ].
Qed.
