Set Implicit Arguments.
Require Import CFLib CFDuplicable.
Require Import LibSeq Cascade_spec.
Require Cascade_ml.

(* -------------------------------------------------------------------------- *)

Theorem nil_spec A:
  @PureCascade A \[] (@epsilon _) nil
    Cascade_ml.nil.
Proof.
  exists (fun xs c => xs = (@nil A) /\ c = Cascade_ml.nil). simpl.
  splits; intros; break; eauto with prefix.
  xcf. xret. xsimpl~.
Qed.

(* -------------------------------------------------------------------------- *)

Theorem cons_spec A:
  forall I seqs,
  forall (x : A) c,
  app Cascade_ml.cons [x c]
    PRE  (c ~> @Cascade A I seqs nil)
    POST (fun c' => c' ~> @Cascade A I (cons x seqs) nil).
Proof.
  (* This is viewed by CFML as a partial application. *)
  intros. xpartial. intros xc. xpull. intros [ ? Hxc ].
  (* There remains to prove that if [c] is a valid cascade
     and if [xc] is the result of the application [cons x c]
     then [xc] is a valid cascade. *)
  xunfold Cascade. xpull. intros S Hdup Hpermitted Hspec.
  (* This requires exhibiting the invariant of the new cascade,
     which must be built on top of the invariant [S] of the
     cascade [c]. Tricky. *)
  xsimpl (fun xs c' =>
    match xs with
    | nil =>
        \[ c' = xc ] \* S nil c
    | y :: ys =>
        \[ y = x ] \* S ys c'
    end
  ).
  { intros [ | y ys ] c'.
    { xpull. intros. subst c'.
      xdup (S nil c). (* TEMPORARY should not be necessary? *)
      xapply~ Hxc.
      xcf. xret.
      xsimpl~. subst. xsimpl~. }
    { xpull. intros. subst y.
      xapply~ Hspec. xsimpl~.
      intros [ | ]; intros; xsimpl~.
      { eauto. }
      { unfold append. xsimpl. (* weird *) }
    }
  }
  { intros [ | y ys ] c'; xpull; intros; subst.
    { xsimpl~. unfold cons. simpl. eauto. }
    { xchange~ Hpermitted. xpull. intros. xsimpl~. unfold cons. simpl. eauto. }
  }
  { intros [ | y ys ] c'; eauto with duplicable. }
  { eauto. }
Qed.
