type 'a vector = {
  mutable default : 'a;
  mutable size : int; (* Logical size *)
  mutable data : 'a array;
}

let length t =
  t.size

let default t =
  t.default

let make size def =
  (* assert (size >= 0); *)
  { default = def;
    size;
    data = Array.make size def;
  }

let init size default f =
  (* assert (size >= 0); *)
  { default;
    size;
    data = Array.init size f;
  }

let get t i =
  (* assert (i >= 0 && i < t.size); *)
  t.data.(i)

let set t i v =
  (* assert (i >= 0 && i < t.size); *)
  t.data.(i) <- v

(* TODO: Replace by Array.blit *)
let transfer src dst nb =
  for i = 0 to nb - 1 do
    dst.(i) <- src.(i);
  done

(* Distinguer le resize_extacly & resize_at_least *)
(* Option de remplir avec default quand on shrink *)
(* set_physical_size *)
(* get_physical_size *)
let resize t new_capa =
  (* assert (t.size <= new_capa); *)
  let old_data = t.data in
  t.data <- Array.make new_capa t.default;
  transfer old_data t.data t.size

let push t x =
  let capa = Array.length t.data in
  if t.size = capa then
    resize t (2 * (capa + 1));
  t.data.(t.size) <- x;
  t.size <- t.size + 1

(* TODO: Return option *)
let pop t =
  (* assert (t.size > 0); *)
  t.size <- t.size - 1;
  let x = t.data.(t.size) in
  let capa = Array.length t.data in
  if (capa > 4) && (t.size <= capa / 4) then
    resize t (capa / 2);
  x

let is_empty t =
  t.size = 0

let copy t =
  { default = t.default;
    size = t.size;
    data = Array.copy t.data;
  }

let to_array t =
  Array.sub t.data 0 (length t)

let of_array default a =
  { default;
    size = Array.length a;
    data = Array.copy a;
  }

let to_list t =
  Array.to_list (to_array t)

let of_list default l =
  { default;
    size = List.length l;
    data = Array.of_list l;
  }

let inplace_sub t start len =
  (* assert (0 <= start <= len && start + len <= t.size); *)
  t.size <- t.size - (len - start);
  t.data <- Array.sub t.data start (t.size - start)

let sub t start len =
  (* assert (0 <= start <= len && start + len <= t.size); *)
  { default = t.default;
    size = t.size - (len - start);
    data = Array.sub t.data start (t.size - start);
  }
