open Printf

exception Error

let work_ref = ref (-1)
let capacity_ref = ref (-1)
let tail_capacity_ref = ref (-1)


(****************************************************************************)

module S = HeadBootChunkedStack

let unit_test () = 
   let capacity = 4 in
   let tail_capacity = capacity * 2 in
   let n = 1000 in
   let s = S.create capacity tail_capacity 0 in
   let debug = true in
   let show s =
      if debug then S.print (fun x -> Printf.printf "%d " x) s (**) in
   for i = 0 to pred n do
      Printf.printf "push %d\n" i;
      S.push i s;
      show s
   done;
   for i = 0 to pred n do
      let x = S.pop s in
      Printf.printf "pop %d = %d\n" i x;
      if (x <> n-1-i) then raise Error;
      show s
   done;
   Printf.printf "ok\n"

let _ =
   let unit = XCmd.mem_flag "unit" in
   if unit then begin
      unit_test();
      exit 0;
   end


(****************************************************************************)


let lifo seq nbitems repeat () () =
   let work = !work_ref in
   let count = ref 0 in
   let sink = ref [] in 
   let do_work () =
      incr count;
      if work > 0 && !count mod work = 0 then begin
         sink := 0::!sink;
      end
      in
   let capacity = !capacity_ref in
   let tail_capacity = !tail_capacity_ref in
   assert (repeat > 0);
   assert (nbitems >= 0);
   let length = nbitems / repeat in
   printf "length %d\n" length;
   if seq = "ocaml_list" then begin

      let r = ref [] in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           r := !a::!r;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = match !r with
              | [] -> assert false
              | x::t -> r := t; x 
              in
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
     done

   end else if seq = "stack" then begin

      let q = Stack.create() in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           Stack.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = Stack.pop q in     
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;


   end else if seq = "inline_array" then begin

      let q = Array.make length 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           q.(i-1) <- !a;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let j = length-i in
           let x = q.(j) in
           q.(j) <- (-1); (* write default to avoid leaks *)
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "inline_array_leaks" then begin

      let q = Array.make length 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           q.(i-1) <- !a;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let j = length-i in
           let x = q.(j) in
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "fixed_size_array" then begin

      let q = FixedCapaStack.create length 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           FixedCapaStack.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = FixedCapaStack.pop q in     
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "fixed_size_array_leaks" then begin

      let q = HardFixedCapaStack.create length 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           HardFixedCapaStack.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = HardFixedCapaStack.pop q in     
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "vector" then begin

      let q = Vector.create 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           Vector.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = Vector.pop q in     
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "tail_boot_stack" then begin

      let q = TailBootChunkedStack.create capacity tail_capacity 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           TailBootChunkedStack.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = TailBootChunkedStack.pop q in     
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "head_boot_stack" then begin

      let q = HeadBootChunkedStack.create capacity tail_capacity 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           HeadBootChunkedStack.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = HeadBootChunkedStack.pop q in     
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "test" then begin

      let q = Test.create capacity tail_capacity 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           Test.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = Test.pop q in     
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "test2" then begin

      let q = Test2.create capacity 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           Test2.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = Test2.pop q in     
           decr a;
           do_work();
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "boot_stack" then begin

      let q = BootChunkedStack.create capacity 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           BootChunkedStack.push (!a) q;
           incr a;
           do_work();
        done;
        for i = 1 to length do
           let x = BootChunkedStack.pop q in     
           decr a;
           do_work();
           assert (x = !a);
        done;
      done;
   

   end else failwith "unsupported seq"


(****************************************************************************)

(*
let lifo_empty seq nbitems repeat () () =
   let capacity = !capacity_ref in
   let tail_capacity = !tail_capacity_ref in
   assert (repeat > 0);
   assert (nbitems >= 0);
   let length = nbitems / repeat in
   printf "length %d\n" length;
   if seq = "ocaml_list" then begin

      let a = ref 0 in

      let r = ref [] in
      for j = 1 to repeat do
        for i = 1 to length do
           r := !a::!r;
           incr a;
        done;
        while !r <> [] do
           let x = match !r with
              | [] -> assert false
              | x::t -> r := t; x 
              in
           decr a;
           if x <> !a then raise Error;
        done;
     done

   end else if seq = "vector" then begin

      let q = Vector.create 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           Vector.push (!a) q;
           incr a;
        done;
        while not (Vector.is_empty q) do
           let x = Vector.pop q in     
           decr a;
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "test" then begin

      let q = Test.create capacity tail_capacity 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           Test.push (!a) q;
           incr a;
        done;
        while not (Test.is_empty q) do
           let x = Test.pop q in     
           decr a;
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "test2" then begin

      let q = Test2.create capacity 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           Test2.push (!a) q;
           incr a;
        done;
        while not (Test2.is_empty q) do
           let x = Test2.pop q in     
           decr a;
           if x <> !a then raise Error;
        done;
      done;

   end else if seq = "boot_stack" then begin

      let q = BootChunkedStack.create capacity 0 in
      let a = ref 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           BootChunkedStack.push (!a) q;
           incr a;
        done;
        while not (BootChunkedStack.is_empty q) do
           let x = BootChunkedStack.pop q in     
           decr a;
           assert (x = !a);
        done;
      done;
   
   end

*)

(*
let lifo_2 seq nbitems repeat () () =
   assert (repeat > 0);
   let length = nbitems / repeat in
   printf "length %d\n" length;
   if seq = "ocaml_list" then begin

      let r = ref [] in
      for j = 1 to repeat do
        for i = 1 to length do
           r := 1::!r;
        done;
        for i = 1 to length do
           match !r with
           | [] -> assert false
           | x::t -> r := t
        done;
     done

   end else if seq = "boot_stack" then begin

      let r = S.create 0 in
      for j = 1 to repeat do
        for i = 1 to length do
           r := S.push 1 !r;
        done;
        for i = 1 to length do
           let (_,t) = S.pop !r in
           r := t
        done;
     done
  end
*)



(****************************************************************************)

exception Unsupported

let measured_run f =
   begin try
      let t1 = Sys.time() in 
      f();
      let t2 = Sys.time() in
      printf "exectime %.3f\n" (t2 -. t1);
   with Unsupported -> printf "exectime NA\n" 
   end

let select choices key =
   try List.assoc key choices
   with Not_found -> failwith (sprintf "not a valid choice: %s\n" key)

let _ =
   let gc_nb = XCmd.parse_or_default_int "gc_nb_32k" 32 in  (* 1Mb *)
   Gc.set {(Gc.get ()) with Gc.minor_heap_size = 32768*gc_nb };
   let seq = XCmd.parse_string "seq" in
   let testname = XCmd.parse_or_default_string "test" "all" in
   let n = XCmd.parse_or_default_int "n" 10000000 in 
   let r = XCmd.parse_or_default_int "r" (-1) in 
   let capacity = XCmd.parse_or_default_int "capacity" 128 in
   capacity_ref := capacity;
   let tail_capacity = XCmd.parse_or_default_int "tail_capacity" (128*128*2) in
   tail_capacity_ref := tail_capacity;
   let work = XCmd.parse_or_default_int "work" 0 in 
   work_ref := work;
   let testnames = [ 
      "lifo", lifo seq n r;
      (*"lifo_empty", lifo seq n r; *)
      ] in
   let test_func = select testnames testname () in
   measured_run test_func



