(* VERSION WITH MEMORY LEAKS *)




(** Arthur Charguéraud *)

(* Remark: using Array.get/set operations insted of notations,
   in order to easily replace those with unsafe array operations *)


(** Representation of fixed-capacity stacks. *)

(*--------------------------------------------------------------------------*)

(** Representation *)

type 'a t = {
  mutable size : int;       (* 0 <= size < capacity *)
  mutable data : 'a array;  (* array of size capacity *) 
  default : 'a; }


(*--------------------------------------------------------------------------*)

(** Builds a new queue *)

let create capacity d = (* take capacity as arg *)
  { size = 0;
    data = Array.make capacity d;
    default = d; }

(** Returns the default element *)

let default q =
  q.default

(** Returns the capacity of the queue *)

let capacity q =
  Array.length q.data 

(** Returns the size of the queue *)

let length q =
  q.size

(** Tests whether the queue is empty *)

let is_empty q =
  q.size = 0

(** Tests whether the queue is full *)

let is_full q =
  q.size = capacity q (* TODO or 128 HARD CODED *)


(*--------------------------------------------------------------------------*)
  
(** Pop an element from the back (assumes non-empty stack) *)

let pop q = 
  assert (q.size > 0);
  q.size <- q.size - 1;
  let x = Array.get q.data q.size in
  (* TODO: deactivated *)
  (* Array.set q.data q.size q.default; (* needed to avoid space leaks *) *)
  x

(** Push an element onto the back (assumes non-full stack) *)

let push x q =
  assert (q.size < capacity q);
  Array.set q.data q.size x;
  q.size <- q.size + 1

(** Top reads the element at the top of the stack, without popping
    (assumes non-empty stack) *)

let top q = 
  Array.get q.data (q.size - 1)


(*--------------------------------------------------------------------------*)

(** Iter *)

let iter f q =
   for k = 0 to pred q.size do
      f (Array.get q.data k);
   done

(** Fold-left *)

let fold_left f a q =
   let acc = ref a in
   for k = 0 to pred q.size do
      acc := f !acc (Array.get q.data k);
   done;
   !acc

(** Fold-right *)

let fold_right f q a =
   let acc = ref a in
   for k = 0 to pred q.size do
      acc := f (Array.get q.data k) !acc;
   done;
   !acc

(** Conversion with lists *)

let to_list q =
   fold_right (fun x acc -> x::acc) q []

