
(** Resizable array.
    Always a power of two.
    Double when full.
    Half when less than a quarter. *)

type 'a t = {
   mutable size : int;
   mutable data : 'a array;
   default : 'a; }

let create d =
 { size = 0;
   data = Array.make 4 d;
   default = d; }

let length t =
   t.size

let is_empty t : bool = 
  t.size = 0

let transfer src dst nb = 
   Array.blit src 0 dst 0 nb
   (* same as:
   for i = 0 to nb - 1 do
      dst.(i) <- src.(i);
   done
   *)

let resize t new_capa =
   let old_data = t.data in
   t.data <- Array.make new_capa t.default;
   transfer old_data t.data t.size

let push x t =
   let capa = Array.length t.data in
   if t.size = capa
      then resize t (2 * capa);
   t.data.(t.size) <- x;
   t.size <- t.size + 1

let pop t = 
   assert (t.size > 0);
   t.size <- t.size - 1;
   let x = t.data.(t.size) in
   t.data.(t.size) <- t.default;
   let capa = Array.length t.data in
   if (capa > 4) && (t.size <= capa/4)
      then resize t (capa / 2);
   x

let get t i =
  assert ((i >= 0) && (i < t.size));
  t.data.(i)

let set t i v =
  assert ((i >= 0) && (i < t.size));
  t.data.(i) <- v








(* --- Demos

let demo_compare (t1:int dyn_array) (t2:int dyn_array) =
  let n1 = length_ t1 in
  let n2 = length_ t2 in
  if n1 <> n2 then
    false
  else begin
    let b = ref true in
    let i = ref 0 in
    while (!b && !i < length_ t1) do
      if (get t1 !i) <> (get t2 !i)
        then b := false;
      incr i
    done;
    !b
  end

let demo_rev_append (t1:'a dyn_array) (t2:'a dyn_array) =
  while not (is_empty t2) do
     let x = pop t2 in
     push t1 x
  done

let demo_seq n =
  let t = make 0 in
  for i = 0 to n - 1 do
    push t i
  done;
  for i = 0 to n - 1 do
    let x = pop t in
    if x <> n - 1 - i then assert false
  done


let demo_pushn t x n =
  for i = 0 to n - 1 do
     push t x;
  done

 *)