
(********************************************************************)
(** Signature for stacks *)

module type StackImpl =
sig

  type 'a t
  (** The type of stacks containing elements of type ['a]. *)

  (* TODO: problème de l'élément par défaut... *)
  (* TODO: question de réexporter une fonction default du coup... *)

  val create : unit -> 'a t
  (** Return a new stack, initially empty. *)

  val clear : 'a t -> unit
  (** Discard all elements from a stack, leaving an empty stack. *)

  val length : 'a t -> int
  (** Return the number of elements in a stack. *)

  val is_empty : 'a t -> bool
  (** Return [true] if the given stack is empty, [false] otherwise. 
     Equivalent to [length s = 0]. *)

  val push : 'a -> 'a t -> unit
  (** [push x s] adds the element [x] at the top of stack [s]. *)

  val pop : 'a t -> 'a
  (** [pop s] removes and returns the topmost element in stack [s]. *)

  val top : 'a t -> 'a
  (** [top s] returns the topmost element in stack [s]. *)

  val iteri : (int -> 'a -> unit) -> 'a t -> unit
  (** [iter f s] applies [f] in turn to all elements of [s],
    from the element at the top of the stack to the element at the
    bottom of the stack. Indices start at 0 for the element at the
    top of the stack. ---TODO: or the bottom? -> iteri_left/right
    The stack itself is unchanged. *)

end


(* TODO: problème de [copy f s] , et de [map]... *)

   (* MAYBE?
   val map : ('a -> 'b) -> 'a t -> 'b t
   (** [map f s] . *)
   *)

   (* MAYBE NOT?
      val copy : 'a t -> 'a t
      (** Return a copy of the given stack. *)
   *)


(* TODO: tabs = 2 or 3 spaces *)




(********************************************************************)
(** Signature for stacks *)

module type S =
sig

  (* TODO: include *)

  type 'a t
  (** The type of stacks containing elements of type ['a]. *)

  exception Empty
  (** Raised when {!Stack.pop} or {!Stack.top} is applied to an empty stack. *)

  val create : unit -> 'a t
  (** Return a new stack, initially empty. *)

  val clear : 'a t -> unit
  (** Discard all elements from a stack, leaving an empty stack. *)

  val length : 'a t -> int
  (** Return the number of elements in a stack. *)

  val is_empty : 'a t -> bool
  (** Return [true] if the given stack is empty, [false] otherwise. 
     Equivalent to [length s = 0]. *)

  val push : 'a -> 'a t -> unit
  (** [push x s] adds the element [x] at the top of stack [s]. *)

  val pop_exn : 'a t -> 'a
  (** [pop s] removes and returns the topmost element in stack [s],
    or raises [Empty] if the stack is empty. *)

  val top_exn : 'a t -> 'a
  (** [top s] returns the topmost element in stack [s],
    or raises [Empty] if the stack is empty. *)

     (* ----TODO: is Empty local to this module or global to vocal *)

  val pop_opt : 'a t -> 'a option
  (** [pop s] removes and returns Some of the topmost element in 
    stack [s], or returns [None] if the stack is empty. *)

  val top_opt : 'a t -> 'a option
  (** [top s] removes and returns Some of the topmost element in 
    stack [s], or returns [None] if the stack is empty. *)

  val iter : ('a -> unit) -> 'a t -> unit
  (** [iter f s] applies [f] in turn to all elements of [s],
    from the element at the top of the stack to the element at the
    bottom of the stack. The stack itself is unchanged. *)

  val iteri : (int -> 'a -> unit) -> 'a t -> unit
  (** [iter f s] applies [f] in turn to all elements of [s],
    from the element at the top of the stack to the element at the
    bottom of the stack. Indices start at 0 for the element at the
    top of the stack. ---TODO: or the bottom? -> iteri_left/right
    The stack itself is unchanged. *)

end

