  (* TODO: choice of names : size vs length *)
  (* TODO: make vs create // ok stick to Ocaml stdlib? *)

type 'a vector = {
  mutable data : 'a array;
  mutable size : int;
  default : 'a;
}

let create size def =
  { data = Array.make size def;
    size = size;
    default = def; }

let init size def f =
  { data = Array.init size f;
    size = size;
    default = def; }

(* internal *)
let transfer src dst nb =
  for i = 0 to nb - 1 do
    dst.(i) <- src.(i);
  done

(* internal *)
(* assert (new_capa >= 0 && new_capa >= size); *)
let realloc t new_capa =
  let new_data = Array.make new_capa t.default in
  transfer t.data new_data t.size;
  t.data <- new_data

let default t =
  t.default

let length t =
  t.size

let is_empty t =
  t.size = 0

  (* new_size may be greater than capacity, in this
  case a reallocation happens;
  if size decreases, default values are written,
  and a reallocation may occur; *)
let set_size t new_size =
  let old_size = t.size in
  t.size <- new_size;
  let old_capa = Array.length t.data in
  if new_size > old_capa then begin
    (* grow the array *)
    let new_capa = max new_size (2 * old_capa) in
    realloc t new_capa;
  end else if new_size < old_capa / 4 then begin
    (* shrink the array *)
    let new_capa = old_capa / 2 in
    realloc t new_capa;
  end else if new_size < old_size then begin
    (* write default values in unused cells *)
    Array.fill t new_size (old_size - new_size) t.default;
  end

(* optimized version of:
Vector.set_size t (length t + 1) *)
let incr_size t =
  let new_size = t.size + 1 in
  t.size <- new_size;
  let old_capa = Array.length t.data in
  if new_size > old_capa then begin
    (* grow the array *)
    let new_capa = max new_size (2 * old_capa) in
    realloc t new_capa;
  end

(* optimized version of:
Vector.set_size t (length t - 1) *)
let decr_size t =
  let new_size = t.size - 1 in
  t.size <- new_size;
  let old_capa = Array.length t.data in
  if new_size < old_capa / 4 then begin
    (* shrink the array *)
    let new_capa = old_capa / 2 in
    realloc t new_capa;
  end else begin
    (* write default values in unused cell *)
    t.data.(new_size) <- t.default
  end

let clear t =
  t.size <- 0;
  realloc t 0

let get t i =
  t.data.(i)

let set t i v =
  t.data.(i) <- v

let iter f t =
  Array.iter f t.data



  (* --------TODO

  let copy t =
  { default = t.default;
  size = t.size;
  data = Array.copy t.data;
  }

  let to_array t =
  let s = length t in
  Array.init s (fun i -> get t i)

  let of_array default a =
  { default;
  size = Array.length a;
  data = Array.copy a;
  }

  *)